Engine.LoadLibrary("rmgen");
Engine.LoadLibrary("rmgen-common");
Engine.LoadLibrary("rmbiome");

TILE_CENTERED_HEIGHT_MAP = true;

setSelectedBiome();

/* REQUIRES Caribe biome selection */

/*
Random Islands Map (Naval) for Aye Pirates Mod Version:	0.9.2
by Happy Concepts
Your feedback to @happyconcepts at ayepirates.com is welcome!
Still Migration in progress from createObjectGroupsDeprecated...

Credits:
This started from islands.js the random gen script by @elexis I *think* :)
Those who've developed 0 A.D. through the years deserve much credit.
This project could not have been done without the OSS developers and
artists who came before :)

*/

const tMainTerrain = g_Terrains.mainTerrain;
const tForestFloor1 = g_Terrains.forestFloor1;
const tForestFloor2 = g_Terrains.forestFloor2;
const tCliff = g_Terrains.cliff;
const tTier1Terrain = g_Terrains.tier1Terrain;
const tTier2Terrain = g_Terrains.tier2Terrain;
const tTier3Terrain = g_Terrains.tier3Terrain;
const tHill = g_Terrains.hill;
const tRoad = g_Terrains.road;
const tRoadWild = g_Terrains.roadWild;
const tTier4Terrain = g_Terrains.tier4Terrain;
const tShore = g_Terrains.shore;
const tWater = g_Terrains.water;
const tCoral = g_Terrains.coral;
const tSandWet = g_Terrains.sandWet;
const tDirt = g_Terrains.dirt;

const oTree1 = g_Gaia.tree1;
const oTree2 = g_Gaia.tree2;
const oTree3 = g_Gaia.tree3;
const oTree4 = g_Gaia.tree4;
const oTree5 = g_Gaia.tree5;
const oFruitBush = g_Gaia.fruitBush;
const oMainHuntableAnimal = g_Gaia.mainHuntableAnimal;
const oFish = g_Gaia.fish;
const oSecondaryHuntableAnimal = g_Gaia.secondaryHuntableAnimal;
const oStoneLarge = g_Gaia.stoneLarge;
const oStoneSmall = g_Gaia.stoneSmall;
const oMetalLarge = g_Gaia.metalLarge;
const oMetalSmall = g_Gaia.metalSmall;
const oWoodTreasure = "gaia/treasure/wood";
const oDock = "skirmish/structures/default_dock";
const oWhale = g_Gaia.whale;
const oShark = g_Gaia.shark;
const oFountain = "structures/fountain_youth";

const aGrass = g_Decoratives.grass;
const aGrassShort = g_Decoratives.grassShort;
const aRockLarge = g_Decoratives.rockLarge;
const aRockMedium = g_Decoratives.rockMedium;
const aBushMedium = g_Decoratives.bushMedium;
const aBushSmall = g_Decoratives.bushSmall;

const pForest1 = [tForestFloor2 + TERRAIN_SEPARATOR + oTree1, tForestFloor2 + TERRAIN_SEPARATOR + oTree2, tForestFloor2];
const pForest2 = [tForestFloor1 + TERRAIN_SEPARATOR + oTree4, tForestFloor1 + TERRAIN_SEPARATOR + oTree5, tForestFloor1];

const heightSeaGround = -3; // -3 is ok for warship
const heightLand = 3;
const heightOffsetBump = 5;
const heightHill = 30;

var g_Map = new RandomMap(heightSeaGround, tWater);

const numPlayers = getNumPlayers();
const mapSize = g_Map.getSize();
const mapCenter = g_Map.getCenter();
const mapBounds = g_Map.getBounds();

var clPlayer = g_Map.createTileClass();
var clHill = g_Map.createTileClass();
var clForest = g_Map.createTileClass();
var clDirt = g_Map.createTileClass();
var clRock = g_Map.createTileClass();
var clMetal = g_Map.createTileClass();
var clFood = g_Map.createTileClass();
var clBaseResource = g_Map.createTileClass();
var clLand = g_Map.createTileClass();

//var startAngle = randomAngle();
var playerIslandRadius = scaleByMapSize(16, 32);

var [playerIDs, playerPosition, playerAngle] = playerPlacementCircle(fractionToTiles(0.35));

if (!isNomad())
{
	g_Map.log("Creating player islands and docks");
	for (let i = 0; i < numPlayers; i++)
	{
		// clump (size, coherence, smoothness, fail, position)
		createArea(
			new ClumpPlacer(diskArea(playerIslandRadius), 0.7, 0.1, Infinity, playerPosition[i]),
			[
				new LayeredPainter([tMainTerrain, tMainTerrain, tMainTerrain], [1, 3]),
				new SmoothElevationPainter(ELEVATION_SET, heightLand + 1, 6),
				new TileClassPainter(clLand),
				new TileClassPainter(clPlayer)
			]);

		let dockLocation = findLocationInDirectionBasedOnHeight(playerPosition[i], mapCenter, -3 , heightLand - 0.5, heightLand);
		g_Map.placeEntityPassable(oDock, playerIDs[i], dockLocation, playerAngle[i] + Math.PI);
	}
}

g_Map.log("Creating big islands");
createAreas(
	new ChainPlacer(
		Math.floor(scaleByMapSize(6, 10)),
		Math.floor(scaleByMapSize(10, 16)),
		Math.floor(scaleByMapSize(30, 64)),
		0.07),
	[
		new TerrainPainter(tMainTerrain),
		new SmoothElevationPainter(ELEVATION_SET, heightLand+1, 6),
		new TileClassPainter(clLand)
	],
	avoidClasses(clLand, scaleByMapSize(3, 8)),
	scaleByMapSize(6, 20));

g_Map.log("Creating small islands");
createAreas(
	new ChainPlacer(
		Math.floor(scaleByMapSize(3, 7)),
		Math.floor(scaleByMapSize(7, 10)),
		Math.floor(scaleByMapSize(16, 40)),
		0.07),
	[
		new TerrainPainter(tMainTerrain),
		new SmoothElevationPainter(ELEVATION_SET, heightLand-1, 5),
		new TileClassPainter(clLand)
	],
	avoidClasses(clLand, scaleByMapSize(3, 10)),
	scaleByMapSize(6, 58));

paintTerrainBasedOnHeight(1, 3, 0, tShore);
//paintTerrainBasedOnHeight(-16, 1, 2, tWater);
paintTerrainBasedOnHeight(-16, 0, 2, tSandWet);

placePlayerBases({
	"PlayerPlacement": [playerIDs, playerPosition],
	// PlayerTileClass marked above
	"BaseResourceClass": clBaseResource,
	"Walls": "towers",
	"CityPatch": {
		"radius": playerIslandRadius / 3,
		"outerTerrain": tRoadWild,
		"innerTerrain": tDirt
	},
	"Chicken": {
	},
	"Berries": {
		"template": oFruitBush,
		"count": 3
	},
	"Mines": {
		"types": [
			{ "template": oMetalLarge }
	/*		,{ "template": oStoneLarge }  */
			]
	},
	"Treasures": {
		"types": [
			{
				"template": oWoodTreasure,
				"count": 4
			}
		]
	},
	"Trees": {
		"template": oTree1,
		"count": 8
	},
	"Decoratives": {
		"template": aGrassShort
	}
});

g_Map.log("Creating bumps");
createAreas(
	new ClumpPlacer(scaleByMapSize(20, 50), 0.3, 0.06, Infinity),
	new SmoothElevationPainter(ELEVATION_MODIFY, heightOffsetBump, 2),
	//[avoidClasses(clPlayer, 1)],
	[ stayClasses(clLand, 1)],
	//[avoidClasses(clPlayer, 1), stayClasses(clLand, 3)],
	scaleByMapSize(20, 100));

	 /*
	  cloned creating bumps ... If we could make them longer and snakier and 1m under water ppl may battle on them
	 */
	g_Map.log("Creating islands");
	createAreas(
		new ClumpPlacer(scaleByMapSize(20, 50), -0.2, 0.06, Infinity),
		new SmoothElevationPainter(ELEVATION_MODIFY, heightOffsetBump - 1, 2),
//		new SmoothElevationPainter(ELEVATION_MODIFY, heightOffsetBump, 2),
		[avoidClasses(clLand, 6)],
		//[avoidClasses(clPlayer, 1), stayClasses(clLand, 3)],
		scaleByMapSize(12, 80));

g_Map.log("Creating deeper water");
// -8
// ChainPlacer(minRadius, maxRadius, numCircles, failFraction = 0, centerPosition = undefined, maxDistance = 0, queue = [])
createAreas(
	new ChainPlacer(3, Math.floor(scaleByMapSize(8, 16)), Math.floor(scaleByMapSize(10, 30)), 0.3),
	[
				new LayeredPainter([tCoral, tWater], 2),
				new SmoothElevationPainter(ELEVATION_SET, -6, 0),
				new TileClassPainter(clHill)
	],
	//[avoidClasses(clPlayer, 2, clHill, 15), stayClasses(clLand, 0)],
			[avoidClasses(clLand, 0)],
			scaleByMapSize(10, 46));


g_Map.log("Creating hills");
createAreas(
	new ChainPlacer(1, Math.floor(scaleByMapSize(3, 10)), Math.floor(scaleByMapSize(18, 44)), 0.9),
	[
		new LayeredPainter([tCliff, tHill], [2]),
		new SmoothElevationPainter(ELEVATION_SET, heightHill, 4),
		new TileClassPainter(clHill)
	],
	[avoidClasses(clPlayer, 0, clHill, 4), stayClasses(clLand, 0)],
	scaleByMapSize(12, 26));



g_Map.log("Creating forests");
var [forestTrees, stragglerTrees] = getTreeCounts(...rBiomeTreeCount(1));
var types = [
	[[tForestFloor2, tMainTerrain, pForest1], [tForestFloor2, pForest1]],
	[[tForestFloor1, tMainTerrain, pForest2], [tForestFloor1, pForest2]]
];

if (currentBiome() != "generic/savanna")
{
	var size = forestTrees / (scaleByMapSize(4,8) * numPlayers);
	var num = Math.floor(size / types.length);
	for (let type of types)
		createAreas(
			new ChainPlacer(1, Math.floor(scaleByMapSize(3, 5)), forestTrees / (num * Math.floor(scaleByMapSize(2, 5))), 0.5),
			[
				new LayeredPainter(type, [2]),
				new TileClassPainter(clForest)
			],
			[avoidClasses(clPlayer, 0, clForest, 10, clHill, 0), stayClasses(clLand, 6)],
			num);
}


Engine.SetProgress(55);

g_Map.log("Creating stone mines");
var group = new SimpleGroup([new SimpleObject(oStoneSmall, 0, 2, 0, 4, 0, 2 * Math.PI, 1), new SimpleObject(oStoneLarge, 1, 1, 0, 4, 0, 2 * Math.PI, 4)], true, clRock);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clForest, 1, clPlayer, 0, clRock, 6, clHill, 1), stayClasses(clLand, 5)],
	scaleByMapSize(4,16), 100
);

g_Map.log("Creating small stone quarries");
group = new SimpleGroup([new SimpleObject(oStoneSmall, 2,5, 1,3)], true, clRock);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clForest, 0, clPlayer, 0, clRock, 0, clHill, 0), stayClasses(clLand, 3)],
	scaleByMapSize(4,16), 100
);

g_Map.log("Creating metal mines");
group = new SimpleGroup([new SimpleObject(oMetalLarge, 4,5, 1,4)], true, clMetal);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clForest, 0, clPlayer, 0, clRock, 0, clHill, 0), stayClasses(clLand, 3)],
	scaleByMapSize(4,16), 100
);

Engine.SetProgress(65);



g_Map.log("Creating small decorative rocks");
group = new SimpleGroup(
	[new SimpleObject(aRockMedium, 1,3, 0,1)],
	true
);
createObjectGroupsDeprecated(
	group, 0,
	[avoidClasses(clForest, 0, clPlayer, 0, clHill, 0), stayClasses(clLand, 5)],
	scaleByMapSize(16, 262), 50
);

g_Map.log("Creating large decorative rocks");
group = new SimpleGroup(
	[new SimpleObject(aRockLarge, 1,2, 0,1), new SimpleObject(aRockMedium, 1,3, 0,2)],
	true
);
createObjectGroupsDeprecated(
	group, 0,
	[avoidClasses(clForest, 0, clPlayer, 0, clHill, 0), stayClasses(clLand, 5)],
	scaleByMapSize(8, 131), 50
);

Engine.SetProgress(72);
// group is (object, min, max, min distance, max disntace)
g_Map.log("Creating whales");
group = new SimpleGroup(
	[new SimpleObject(oWhale, 1, 2, 0, 2)],
	true, clFood
);
// params group,2,locationish,amount,retry factor
createObjectGroups(group, 0,
	[avoidClasses(clLand, 6)],
	(1 * numPlayers), 8
);

g_Map.log("Creating fountain");
group = new SimpleGroup(
	[new SimpleObject(oFountain, 1, 1, 0, 2)],
	true, clFood
);
//  paramsseem to be group,2,locationish,amount,retry factor
createObjectGroups(group, 0,
	[avoidClasses(clPlayer, 0, clFood, 0, clHill, 3), stayClasses(clLand, 6)],
	1, 10
);

g_Map.log("Creating boar");
group = new SimpleGroup(
	[new SimpleObject(oMainHuntableAnimal,3,4,0,4)],
	true, clFood
);
createObjectGroups(group, 0,
	[avoidClasses(clHill, 0, clPlayer, 0, clFood, 0), stayClasses(clLand, 10)],
	3 * numPlayers, 20
);

Engine.SetProgress(75);

g_Map.log("Creating sheep");
group = new SimpleGroup(
	[new SimpleObject(oSecondaryHuntableAnimal, 2,3, 0,2)],
	true, clFood
);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clForest, 0, clPlayer, 0, clHill, 1, clFood, 10), stayClasses(clLand, 6)],
	3 * numPlayers, 50
);

g_Map.log("Creating fruit bushes");
group = new SimpleGroup(
	[new SimpleObject(oFruitBush, 5,7, 0,4)],
	true, clFood
);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clForest, 0, clPlayer, 6, clHill, 1, clFood, 10), stayClasses(clLand, 8)],
	randIntInclusive(1, 4) * numPlayers + 2, 50
);

g_Map.log("Creating fish");
group = new SimpleGroup(
	[new SimpleObject(oFish, 2,3, 0,2)],
	true, clFood
);
createObjectGroupsDeprecated(group, 0,
	avoidClasses(clLand, 4, clForest, 2, clPlayer, 2, clHill, 2, clFood, 16),
	50 * numPlayers, 60
);

Engine.SetProgress(85);

createStragglerTrees(
	[oTree1, oTree5, oTree4, oTree3],
	[avoidClasses(clForest, 1, clHill, 1, clPlayer, 0, clMetal, 6, clRock, 6), stayClasses(clLand, 6)],
	clForest,
	stragglerTrees);

var planetm = 1;
if (currentBiome() == "generic/tropic")
	planetm = 8;

g_Map.log("Creating small grass tufts");
group = new SimpleGroup(
	[new SimpleObject(aGrassShort, 1,2, 0,1, -Math.PI / 8, Math.PI / 8)]
);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clHill, 2, clPlayer, 2, clDirt, 0), stayClasses(clLand, 6)],
	planetm * scaleByMapSize(13, 200)
);

Engine.SetProgress(90);

g_Map.log("Creating large grass tufts");
group = new SimpleGroup(
	[new SimpleObject(aGrass, 2,4, 0,1.8, -Math.PI / 8, Math.PI / 8), new SimpleObject(aGrassShort, 3,6, 1.2,2.5, -Math.PI / 8, Math.PI / 8)]
);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clHill, 2, clPlayer, 1, clDirt, 1, clForest, 0), stayClasses(clLand, 5)],
	planetm * scaleByMapSize(13, 200)
);

Engine.SetProgress(95);

g_Map.log("Creating bushes");
group = new SimpleGroup(
	[new SimpleObject(aBushMedium, 1,2, 0,2), new SimpleObject(aBushSmall, 2,4, 0,2)]
);
createObjectGroupsDeprecated(group, 0,
	[avoidClasses(clHill, 1, clPlayer, 1, clDirt, 1), stayClasses(clLand, 6)],
	planetm * scaleByMapSize(13, 200), 50
);

placePlayersNomad(clPlayer, [stayClasses(clLand, 4), avoidClasses(clHill, 2, clForest, 1, clMetal, 4, clRock, 4, clFood, 2)]);
setSkySet(pickRandom(["sunny"]));
setSunRotation(randomAngle());
setSunElevation(randFloat(1/5, 1/3) * Math.PI);
//setWaterWaviness(2);
//setWaterHeight(18.5);
g_Map.ExportMap();
